package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.function.CharProvider;
import dev.pfaff.altargui.render.objects.TextRenderObject;
import dev.pfaff.altargui.text.Font;
import dev.pfaff.altargui.text.MinecraftBitmapFont;
import dev.pfaff.altargui.text.StyleFlags;
import dev.pfaff.altargui.texture.LoadableTexture;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.util.BasicFuture;
import dev.pfaff.altargui.util.LinkedNode;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.listenable.Value;
import dev.pfaff.altargui.util.WidgetStateSignalManager;
import dev.pfaff.altargui.widgets.state.Painter;
import org.jetbrains.annotations.NotNull;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.invoke.MethodHandles;
import java.util.function.Consumer;
import java.util.function.IntUnaryOperator;

import static dev.pfaff.altargui.text.StyleFlags.BOLD;
import static dev.pfaff.altargui.text.StyleFlags.isBold;

/**
 * A widget that displays a segment of uniformly-styled text.
 *
 * @param color text colorRgba (ARGB)
 * @param font the value may be null while the font is loading
 * @param flags {@link StyleFlags}
 */
public record TextFragmentWidget(CharProvider characters, @NonNull Value<? extends @Nullable Font> font, int color, int flags, Value<Float> fontSize) implements TextWidget, LeafWidget {
	/**
	 * @deprecated Use {@link #TextFragmentWidget(CharProvider, Value, Color, int, Value)} instead.
	 */
	@Deprecated
	public TextFragmentWidget {
	}

	public TextFragmentWidget(CharProvider characters, @NonNull Value<? extends @Nullable Font> font, Color color, int flags, Value<Float> fontSize) {
		this(characters, font, color.rgba(), flags, fontSize);
	}

	public TextFragmentWidget(CharProvider characters, @NonNull Value<? extends @Nullable Font> font) {
		// TODO: this is too bright. Find the correct default colorRgba value.
		this(characters, font, DEFAULT_COLOR, StyleFlags.REGULAR, TextWidget.DEFAULT_SCALED_FONT_SIZE_LISTENABLE);
	}

	public TextFragmentWidget(char character, @NonNull Value<? extends @Nullable Font> font) {
		this(CharProvider.of(character), font);
	}

	public TextFragmentWidget(String string, @NonNull Value<? extends @Nullable Font> font) {
		this(CharProvider.of(string), font);
	}

	@Override
	public TextWidget withFont(Value<? extends @Nullable Font> font) {
		return new TextFragmentWidget(characters, font, color, flags, fontSize);
	}

	@Override
	public TextWidget withFontSize(Value<Float> fontSize) {
		return new TextFragmentWidget(characters, font, color, flags, fontSize);
	}

	public TextFragmentWidget withColor(Color color) {
		return new TextFragmentWidget(characters, font, color, flags, fontSize);
	}

	@Override
	public TextFragmentWidget mapFlags(IntUnaryOperator modifier) {
		return new TextFragmentWidget(characters, font, color, modifier.applyAsInt(flags), fontSize);
	}

	public TextFragmentWidget withFlags(int flags) {
		return (TextFragmentWidget) TextWidget.super.withFlags(flags);
	}

	public TextFragmentWidget withBold(boolean bold) {
		return (TextFragmentWidget) TextWidget.super.withBold(bold);
	}

	public TextFragmentWidget withItalic(boolean italic) {
		return (TextFragmentWidget) TextWidget.super.withItalic(italic);
	}

	public TextFragmentWidget withStrike(boolean strike) {
		return (TextFragmentWidget) TextWidget.super.withStrike(strike);
	}

	public TextFragmentWidget withUnderline(boolean underline) {
		return (TextFragmentWidget) TextWidget.super.withUnderline(underline);
	}

	@Override
	public @NonNull WidgetInstance<TextFragmentWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends LeafWidget.AbstractState<TextFragmentWidget> implements Painter {
		private static final int LAYOUT_DEPENDENCY_FLAGS = BOLD;

		private static final WidgetStateSignalManager<State, Value<? extends @Nullable Font>> FONT_SIGNAL_MANAGER = new WidgetStateSignalManager<>(
				MethodHandles.lookup(),
				"fontSignal", (Class<Value<? extends @Nullable Font>>) (Class) Value.class,
				"fontListener",
				(state, signal) -> signal.listenAndInvoke(font -> {
					state.font = font;
					state.texDirty = true;
					state.charsDirty = true;
					state.markNeedsStencilLayout();
				})
		);
		private static final WidgetStateSignalManager<State, Value<@NotNull Float>> FONT_SIZE_SIGNAL_MANAGER = new WidgetStateSignalManager<>(
				MethodHandles.lookup(),
				"fontSizeSignal", (Class<Value<@NotNull Float>>) (Class) Value.class,
				"fontSizeListener",
				(state, signal) -> signal.listenAndInvoke(size -> {
					state.fontSize = size;
					state.charsDirty = true;
					state.markNeedsStencilLayout();
				})
		);

		private final TextRenderObject renderObject = new TextRenderObject(0);
		private CharProvider characters;
		private Value<? extends Font> fontSignal;
		private int color;
		private int flags;
		private Value<Float> fontSizeSignal;

		private @Nullable ListenerDisposer fontListener;
		private @Nullable ListenerDisposer fontSizeListener;

		private @Nullable Font font;
		private float fontSize = 8f;

		private boolean charsDirty = false;
		private boolean texDirty = false;

		@Nullable
		private LinkedNode<LoadableTexture> textures;

		@Nullable
		private LinkedNode<BasicFuture<Void>> textureLoadFutures;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, TextFragmentWidget widget, boolean init) {
			// this is here for a reason
			boolean texDirty = characters == null || widget.characters.length() != characters.length();
			boolean layoutDirty = false;
			layoutDirty |= texDirty || !widget.characters.equals(characters);
			layoutDirty |= (widget.flags & LAYOUT_DEPENDENCY_FLAGS) != (flags & LAYOUT_DEPENDENCY_FLAGS);
			boolean charsDirty = layoutDirty;
			charsDirty |= widget.flags != flags;
			charsDirty |= widget.color != color;
			if (texDirty) {
				this.texDirty = true;
			}
			if (charsDirty) {
				this.charsDirty = true;
				this.color = widget.color;
				this.characters = widget.characters;
				this.flags = widget.flags;
				markNeedsStencilLayout();
			}
//			if (layoutDirty) markNeedsLayout();
			FONT_SIZE_SIGNAL_MANAGER.configure(this, widget.fontSize);
			FONT_SIGNAL_MANAGER.configure(this, widget.font);
		}

		@Override
		protected void onMount() {
			super.onMount();

			FONT_SIZE_SIGNAL_MANAGER.onMount(this);
			FONT_SIGNAL_MANAGER.onMount(this);
		}

		@Override
		protected void onUnmount() {
			FONT_SIZE_SIGNAL_MANAGER.onUnmount(this);
			FONT_SIGNAL_MANAGER.onUnmount(this);
			if (this.textureLoadFutures != null) {
				this.textureLoadFutures.forEach(BasicFuture::cancel);
				this.textureLoadFutures = null;
			}
			super.onUnmount();
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			// TODO: optimize
			if (charsDirty) {
				if (this.font == null) return;
				charsDirty = false;
				var flags = this.flags;
				var fontSize = this.fontSize;
				var boldAdvance = isBold(flags) ? MinecraftBitmapFont.BOLD_ADVANCE * fontSize : 0.0f;
				var regularSpacing = MinecraftBitmapFont.SPACING * fontSize;
				var spacing = regularSpacing + boldAdvance;
				var renderObject = this.renderObject;
				var color = this.color;
				var font = this.font;
				var characters = this.characters;
				var provider = new TextRenderObject.Provider() {
					@Nullable
					private LinkedNode<LoadableTexture> textures;
					private float x = 0f;

					@Override
					public void accept(TextRenderObject.Sink sink) {
						characters.provide(c -> {
							var glyph = font.getGlyph(c);
							this.textures = new LinkedNode<>(glyph.texture(), this.textures);
							var minX = glyph.minX();
							var maxX = glyph.maxX();
							this.x -= minX * fontSize;
							// FIXME: all characters are currently offset by +1 pixels on the X-axis.
							sink.accept((byte) c, color, flags, this.x, 0f, fontSize);
							this.x += maxX * fontSize + spacing;
						});
					}
				};
//				Logger.log(() -> STR."Updating characters with \{characters.length()} characters from '\{characters}'");
				renderObject.updateCharacters(characters.length(), provider);
				this.textures = provider.textures;
				if (this.textureLoadFutures != null) {
					this.textureLoadFutures.forEach(BasicFuture::cancel);
				}
				// remove trailing spacing
				if (provider.x != 0f) provider.x -= regularSpacing;
//			Logger.log(() -> STR."TextFragmentWidget of \{characters} goes from \{initX},\{y} to \{consumer.x},\{y + fontSize}");
				setComputedSize(provider.x, fontSize);
			}
		}

		@Override
		protected void computeTraceLayout() {
			this.renderObject.updatePosition(allocatedX(), allocatedY());
		}

		private static GuiException throwNotEnoughSpaceException(boolean canOverflowX, boolean canOverflowY, float requiredWidth, float requiredHeight, float availableWidth, float availableHeight) {
			var sb = new StringBuilder("Not enough available space for text fragment: ");
			boolean overflowsX = !canOverflowX && requiredWidth > availableWidth;
			boolean overflowsY = !canOverflowY && requiredHeight > availableHeight;
			if (overflowsX) {
				sb.append("required width ")
						.append(requiredWidth)
						.append(" overflows X-axis by ")
						.append(requiredWidth - availableWidth);
			}
			if (overflowsX && overflowsY) {
				sb.append(" and ");
			}
			if (overflowsY) {
				sb.append("required height ")
						.append(requiredHeight)
						.append(" overflows Y-axis by ")
						.append(requiredHeight - availableHeight);
			}
			sb.append(".");
			return new GuiException(sb.toString());
		}

		@Override
		public void mount(MountContext context) {
			super.mount(context);
		}

		@Override
		public void paint(long time, Canvas canvas) {
			if (this.font != null) {
				if (this.texDirty) {
					this.texDirty = false;
					assert this.font != null;
					if (this.textures != null) {
						var consumer = new Consumer<LoadableTexture>() {
							@Nullable
							private LinkedNode<BasicFuture<Void>> futures = null;
							private int i = 0;

							@Override
							public void accept(LoadableTexture texture) {
								this.futures = new LinkedNode<>(canvas.texAlloc().alloc(texture)
										.cancelOnInterrupt()
										.thenAccept(handle -> {
											if (handle.textureWidth != handle.textureHeight) {
												throw new IllegalArgumentException(
														"Invalid font page texture: width != height");
											}
											// TODO: minus one, add back in shader, allowing up to 256 instead of 255.
											var extent = Math.divideExact(handle.textureWidth, 16);
											if (Byte.toUnsignedInt((byte)extent) != extent) {
												throw new IllegalArgumentException("Invalid font page texture: glyph size is greater than 255 pixels");
											}
											State.this.renderObject.updateTex(this.i++,
													handle.unitU(),
													handle.unitV(),
													(byte) extent);
										})
										.uncatch(), this.futures);
							}
						};
						this.textures.forEach(consumer);
						this.textures = null;
						this.textureLoadFutures = consumer.futures;
					} else {
						this.textureLoadFutures = null;
					}
				}
				canvas.upload(this.renderObject);
			}
		}
	}
}
